# -*- coding: utf-8 -*-
from typing import Tuple
import numpy as np
import cv2 as cv


def draw_rows(
    image: np.ndarray,
    rows: np.ndarray,
    line_color: Tuple[int, int, int] = (255, 0, 0),
    line_width: int = 2
):
    """Draw crop rows lines in a crop field image.

    Parameters
    ----------
    image : array-like of shape = {[m, n, 3], [m, n]} and ``dtype=np.uint8``
        Color or grayscale crop field image.
    rows : array-like of shape = [r, c, 2] and ``dtype=np.int32``
        Crop rows lines, where each element is crop row line represented by an
        array of `c` pairs of indexes in the form `(row, column)`. Each crop
        row line is a continuous polygonal line.
    line_color : tuple of ints with length = 3, optional, (default=(255, 0, 0))
        Color of drawn lines, in BGR format.
    line_width : int, optional, (default=2)
        Width of drawn lines.

    Returns
    -------
    image : array-like of shape = {[m, n, 3], [m, n]} and ``dtype=np.uint8``
        Image with drawn crop rows lines. Input `image` is not modified, drawing
        is done in a copy of it.

    """
    if image.ndim == 2:
        image = cv.cvtColor(image, cv.COLOR_GRAY2BGR)
    else:
        image = image.copy()

    for k in range(rows.shape[0]):
        for n in range(rows.shape[1] - 1):
            p1 = (rows[k, n, 1], rows[k, n, 0])
            p2 = (rows[k, n + 1, 1], rows[k, n + 1, 0])
            cv.line(image, p1, p2, line_color, line_width)

    return image


def poly_mask(poly: np.ndarray, shape: tuple):
    roi_mask = np.zeros(shape, dtype=np.uint8)
    cv.fillConvexPoly(roi_mask, poly, 255)
    return roi_mask


def array_image(
    values: np.ndarray,
    colormap: int = None,
    full_scale: bool = True
):
    if values.dtype != np.uint8 or full_scale:
        value_max = float(np.max(values))
        value_min = float(np.min(values))
        if value_max <= 255 and value_min >= 0 and not full_scale:
            values = np.uint8(values)
        elif not full_scale or value_min == value_max:
            values = np.clip(values, 0, 255, dtype=np.uint8, casting='unsafe')
        else:
            m = 255.0 / (value_max - value_min)
            values = np.uint8(m * (values - value_min))

    if colormap is not None:
        values = cv.applyColorMap(values, colormap)

    return values


# Rotation is CCW with rotation center in upper left corner
def rotate_image(
    image: np.ndarray,
    angle: float,
    interp: int = cv.INTER_NEAREST,
    border_mode: int = cv.BORDER_TRANSPARENT,
    border_value: int = 0
):
    (h, w) = image.shape[0:2]
    matrix = transform_matrix(angle)
    corners = np.array([[0, 0, 1], [w, 0, 1], [w, h, 1], [0, h, 1]])
    corners = np.int32(np.dot(matrix, corners.T)).T[:, 0:-1]
    x, y, w, h = cv.boundingRect(corners.reshape(1, -1, 2))
    matrix[0:2, -1] = [-x, -y]
    image = cv.warpAffine(
        image,
        matrix[0:2, :],
        (w, h),
        flags=interp,
        borderMode=border_mode,
        borderValue=border_value
    )
    return image, matrix


def trim_image(img, box):
    x1, y1, w, h = box
    x2, y2 = x1 + w, y1 + h
    return img[y1:y2, x1:x2]


def trim_poly(poly, box):
    x1, y1, w, h = box
    x2, y2 = x1 + w, y1 + h
    return np.minimum(np.maximum(poly, [x1, y1]), [x2, y2])


# Rotation is CCW
def transform_matrix(
    angle: float = 0,
    translate: tuple = (0, 0),
    scale: tuple = (1.0, 1.0)
):
    s, c = np.sin(angle), np.cos(angle)
    sx, sy = scale[0], scale[1]
    dx, dy = translate[0], translate[1]
    matrix = np.array([[sx*c, sy*s, dx], [-sx*s, sy*c, dy], [0, 0, 1]])
    return matrix


# -*- coding: utf-8 -*-
import time
from math import pi
from typing import List, Tuple

import cv2 as cv
import numpy as np
from sklearn.tree import DecisionTreeClassifier

from .utils import rotate_image


def segment_weeds(
    image: np.ndarray,
    veg_mask: np.ndarray,
    crop_rows: np.ndarray,
    model: DecisionTreeClassifier = None
):
    """Segment weed pixels of a crop field image.

    Parameters
    ----------
    image : array-like of shape = [m, n, 3] and ``dtype=np.uint8``
        The crop field image to be segmented.
    veg_mask : array-like of shape = [m, n] and ``dtype=np.uint8``
        Vegetation mask image, where non zero and zero elements represents
        vegetation and background pixels, respectively.
    crop_rows : array-like of shape = [r, c, 2] and ``dtype=np.int32``
        Crop rows lines, where each element is crop row line represented by an
        array of `c` pairs of indexes in the form `(row, column)`. Each crop
        row line is a continuous polygonal line.
    model : sklearn.tree.DecisionTreeClassifier or None, optional (default=None)
        Decision tree-based model to classify vegetation pixels into "crop" and
        "weed" classes. Crop and weed pixels must be labeled as `0` and `1`,
        respectively.

    Returns
    -------
    weed_mask : array-like of shape = [m, n] and ``dtype=np.uint8``
        Weed mask image, where non zero and zero elements represents weed and
        background pixels, respectively.

    See also
    --------
    detect_rows, classification_model, segment_vegetation

    """
    veg_mask_bool = veg_mask > 0

    if model is None:
        dist_map = _row_distance_map(veg_mask, crop_rows)
        dist_thr = np.mean(dist_map[veg_mask_bool])
        pixels_crop = image[(dist_map < dist_thr) & (dist_map > 0)]
        pixels_weed = image[dist_map > 2 * dist_thr]
        model = classification_model(pixels_crop, pixels_weed)

    predictions = model.predict(image[veg_mask_bool])
    values = np.zeros((predictions.size, 1), np.uint8)
    values[predictions == 1] = 255
    weed_mask = np.zeros_like(veg_mask)
    weed_mask[veg_mask_bool] = values.reshape(-1)

    return weed_mask


def _row_distance_map(veg_mask: np.ndarray, rows: np.ndarray):
    """Compute distances between vegetation pixels and crop row lines.

    Parameters
    ----------
    veg_mask : array-like of shape = [m, n] and ``dtype=np.uint8``
        Vegetation mask image, where non zero and zero elements represents
        vegetation and background pixels, respectively.
    rows : array-like of shape = [r, c, 2] and ``dtype=np.int``
        Crop rows lines, where each element is crop row line represented by
        an array of `c` pairs of indexes in the form `(row, column)`. Each crop
        row line is a continuous polygonal line.

    Returns
    -------
    distance_map : array-like of shape = [m, n] and ``dtype=np.int32``
        An array of the same shape as the input `veg_mask`, where each
        element that correspond to a non zero element in `veg_mask` is non zero,
        represents the distance in pixels from that pixel to the closest crop
        row line.

    """
    distance_map = np.zeros(veg_mask.shape, np.int32)
    h, w = veg_mask.shape[:2]
    n_rows, n_cols = rows.shape[:2]

    for k in range(n_rows):
        for n in range(n_cols):
            if k == 0:
                r1 = 0
            elif n == n_cols - 1:
                r1 = int((rows[k, n, 0] + rows[k - 1, n, 0]) / 2)
            else:
                r1 = int((
                    rows[k, n, 0] + rows[k, n + 1, 0] +
                    rows[k - 1, n, 0] + rows[k - 1, n + 1, 0]
                ) / 4)

            if k == n_rows - 1:
                r2 = veg_mask.shape[0]
            elif n == n_cols - 1:
                r2 = int((rows[k, n, 0] + rows[k + 1, n, 0]) / 2)
            else:
                r2 = int((
                    rows[k, n, 0] + rows[k, n + 1, 0] +
                    rows[k + 1, n, 0] + rows[k + 1, n + 1, 0]
                ) / 4)

            c1 = rows[k, n, 1]

            if n == n_cols - 1:
                c2 = veg_mask.shape[1]
                value = rows[k, n, 0]
            else:
                c2 = rows[k, n + 1, 1]
                value = int((rows[k, n, 0] + rows[k, n + 1, 0]) / 2)

            distance_map[r1:r2, c1:c2] = value

    distance_map = np.abs(
        distance_map - np.arange(0, h).reshape((-1, 1)).repeat(w, axis=1)
    )
    # cv.namedWindow('test image', cv.WINDOW_NORMAL)
    # cv.resizeWindow('test image', 900, 576)
    # ci = array_image(distance_map)
    # ci = cv.applyColorMap(ci, cv.COLORMAP_HSV)
    # cv.imshow('test image', ci)
    # ret = cv.waitKey(0)
    return distance_map


def classification_model(
    pixels_crop: np.ndarray,
    pixels_weed: np.ndarray
):
    """Build a pixel classification model based on decision tree.

    Crop pixels are labeled as `0`, while weed pixels are labeled as `1`.

    Parameters
    ----------
    pixels_crop : array-like of shape = [p, 3] and ``dtype=np.uint8``
        Crop pixels data set, where each element represents a pixel of crop
        image regions.
    pixels_weed : array-like of shape = [q, 3] and ``dtype=np.uint8``
        Weeds pixels data set, where each element represents a pixel of weed
        image regions.

    Returns
    -------
    model : DecisionTreeClassifier object
        The DecisionTreeClassifier object. Please refer to
        ``help(sklearn.tree.DecisionTreeClassifier)`` for attributes of
        DecisionTreeClassifier object and basic usage of these attributes.

    """
    n_crops, n_weeds = pixels_crop.shape[0], pixels_weed.shape[0]
    inputs = np.vstack((pixels_crop, pixels_weed))
    outputs = np.ones((n_crops + n_weeds, 1), np.uint8)
    outputs[0:n_crops] = 0
    model = DecisionTreeClassifier()
    model.fit(inputs, outputs)
    return model


def fake_weeds(
    image: np.ndarray,
    veg_mask: np.ndarray,
    weed_images: List[np.ndarray],
    n_patches: int = 100,
    patch_density: float = 0.2,
    patch_size: Tuple[int, int] = (100, 100),
    resolution: float = 20
):
    """Add weed patches to a crop field image.

    Parameters
    ----------
    image : array-like of shape = [m, n, 3] and ``dtype=np.uint8``
        The crop field image.
    veg_mask : array-like of shape = [m, n] and ``dtype=np.uint8``
        Vegetation mask image, where non zero and zero elements represents
        vegetation and background pixels, respectively.
    weed_images : list of array-like of shape = [m, n, 4] and ``dtype=np.uint8``
        List of the four-channel weed image samples.
    n_patches : int, optional, (default=100)
        Number of weed patches to add.
    patch_density : float, optional, (default=0.2)
        Weed density inside each added weed patch.
    patch_size : tuple of ints with length = 2, optional, (default=(2, 2))
        Mean size (width x height) in meters of the weed patches
    resolution : float, optional, (default=20)
        Resolution in pixels/meter of the input mask image.


    Returns
    -------
    image_weeded : array-like of shape = [m, n, 3] and ``dtype=np.uint8``
        Crop field image with added weed patches.
    weed_mask : array-like of shape = [m, n] and ``dtype=np.uint8``
        Weed mask image, where non zero and zero elements represents weed and
        background pixels, respectively.

    """

    veg_density_thr = 0.1
    max_iter = 100000
    patch_size_var = 0.9
    rotation_aug = 5

    patch_size = tuple([int(size * resolution) for size in patch_size])

    weed_images_aug = []
    for weed_image in weed_images:
        for n in range(rotation_aug):
            weed_image, _ = rotate_image(
                weed_image,
                np.random.uniform(0, 2 * pi),
                interp=cv.INTER_LINEAR,
                border_mode=cv.BORDER_CONSTANT
            )
            weed_images_aug.append(weed_image)

    weed_images = weed_images_aug

    weed_areas = []
    weed_masks = []
    for weed_image in weed_images:
        alpha = weed_image[:, :, 3]
        mask = np.zeros(weed_image.shape[:2], np.uint8)
        mask[alpha > 250] = 255
        weed_masks.append(mask)
        weed_areas.append(np.count_nonzero(mask))

    n_weeds = len(weed_images)

    weed_mask = np.zeros_like(veg_mask)
    image_weeded = np.copy(image)

    x_max, y_max = image.shape[1] - 1, image.shape[0] - 1
    xw_max, yw_max = int(0.5 * patch_size[0]), int(0.5 * patch_size[1])
    xw_min, yw_min = patch_size_var * xw_max, patch_size_var * yw_max

    np.random.seed(int(time.time()))
    iter_count = 0
    n = 0
    while n < n_patches:
        x = np.random.randint(0, x_max)
        y = np.random.randint(0, y_max)
        wr = np.random.randint(xw_min, xw_max)
        hr = np.random.randint(yw_min, yw_max)
        x1, x2 = max(0, x - wr), min(x + wr + 1, x_max + 1)
        y1, y2 = max(0, y - hr), min(y + hr + 1, y_max + 1)
        if (x1 < 0) or (y1 < 0) or (x2 > x_max) or (y2 > y_max):
            continue
        patch_area = (2 * wr + 1) * (2 * hr + 1)
        veg_density = np.count_nonzero(veg_mask[y1:y2, x1:x2]) / patch_area
        if veg_density > veg_density_thr:
            continue

        max_weed_area = patch_area * patch_density
        weed_area = 0
        while weed_area < max_weed_area:
            weed_ind = np.random.randint(0, n_weeds)
            weed_image = weed_images[weed_ind]
            xw1, yw1 = np.random.randint(x1, x2), np.random.randint(y1, y2)
            xw2, yw2 = xw1 + weed_image.shape[1], yw1 + weed_image.shape[0]
            if (xw2 > x_max) or (yw2 > y_max):
                continue
            alpha = weed_image[:, :, 3] / 255.0
            alpha = np.dstack((alpha, alpha, alpha))
            image_weeded[yw1:yw2, xw1:xw2] = cv.add(
                (1 - alpha) * image[yw1:yw2, xw1:xw2],
                alpha * weed_image[:, :, 0:3]
            )
            weed_mask[yw1:yw2, xw1:xw2] = weed_masks[weed_ind][:]
            weed_area += np.count_nonzero(weed_mask[y1:y2, x1:x2])
            iter_count += 1

        n += 1
        iter_count += 1
        if iter_count > max_iter:
            break

    return image_weeded, weed_mask





# -*- coding: utf-8 -*-
"""A small library for processing images of agricultural fields"""

__version__ = '0.0.1'
__author__ = 'Raikel Bordon'
__contact__ = 'raikelbl@gmail.com'
__license__ = 'MIT'

from . import rows, veget, weeds, utils

__all__ = ('rows', 'veget', 'weeds', 'utils')



# -*- coding: utf-8 -*-
import cv2 as cv
import numpy as np

from .utils import trim_poly, rotate_image, trim_image, transform_matrix


def segment_vegetation(image: np.ndarray, threshold: float = 1):
    """Segment vegetation pixels of a crop field image.

    Parameters
    ----------
    image : array-like of shape = [m, n, 3] and ``dtype=np.uint8``
        The crop field image to be segmented.
    threshold : float, optional, (default=1)
        Threshold segmentation value.

    Returns
    -------
    mask : array-like of shape = [m, n] and ``dtype=np.int32``
        A mask image where non zero and zero elements represents
        vegetation and background pixels in the input image, respectively.

    """

    if image.dtype != np.float:
        image = np.float32(image)
    b, g, r = cv.split(image)
    index = (2 * g - r - b)
    mask = np.zeros(image.shape[:2], np.uint8)
    mask[index > threshold] = 255
    return mask


def norm_image(
    image: np.ndarray,
    roi_poly: np.ndarray = None,
    roi_trim: bool = True,
    rows_direction: float = 0,
    is_mask: bool = False
):
    """Normalize a crop field image.

    A crop field image is normalized by rotating it by and angle equal in
    magnitude to the crop rows direction, but in opposite direction, such that
    in the normalized image the crop rows are horizontally oriented. Optionally,
    the image can be cropped to an specified region-of-interest (ROI).

    Parameters
    ----------
    image : array-like of shape = [m, n, 3] or [m, n] and ``dtype=np.uint8``
        Crop field image. Can be a multi-channel or single-channel image.
    roi_poly : array-like of shape = [r, 1, 2] and ``dtype=np.int32``
        Convex polygon that encloses the ROI of the input image, in the format
        returned by ``help(cv2.contourArea)``.
    roi_trim : bool, optional, (default=True)
        Whether to trim the image to the roi area.
    rows_direction : float, optional, (default=0.0)
        The crop rows mean direction, clock wise, in radians.
    is_mask: bool, optional, (default=False)
        Whether `image` is a binary mask image. This controls the interpolation
        method used internally to transform images. For binary mask images, the
        method used is `cv.INTER_NEAREST`, otherwise the method used is
        `cv.INTER_LINEAR`.

    Returns
    -------
    image : array-like of shape = [m, n, 3] or [m, n] and ``dtype=np.int32``
        An array of the same shape as the input `image`, representing the
        normalized image.
    roi_poly: array-like of shape = [r, 1, 2] and ``dtype=np.int32``
        The modified input `roi_poly`, after normalization
    t_matrix: array-like of shape = [3, 3] and ``dtype=np.float``
        The transformation matrix that make the normalization.

    """

    (h, w) = image.shape[0:2]
    interp = cv.INTER_NEAREST if is_mask else cv.INTER_LINEAR

    if roi_poly is not None:
        roi_poly = np.array(roi_poly, np.int32)
        roi_poly = np.reshape(roi_poly, (-1, 1, 2))
        roi_poly = trim_poly(roi_poly, (0, 0, w, h))

    image, t_matrix = rotate_image(image, rows_direction, interp=interp)
    (h, w) = image.shape[0:2]

    if roi_poly is not None:
        roi_poly = cv.transform(roi_poly, t_matrix[0:2, :])
        roi_poly = trim_poly(roi_poly, (0, 0, w, h))

    if roi_trim and roi_poly is not None:
        roi_box = cv.boundingRect(roi_poly)
        image = trim_image(image, roi_box)
        box_corner = (roi_box[0], roi_box[1])
        roi_poly -= [box_corner]
        t_matrix = np.dot(
            np.linalg.inv(t_matrix),
            transform_matrix(translate=box_corner)
        )

    roi_mask = np.zeros(image.shape[0:2], np.uint8)
    cv.fillConvexPoly(roi_mask, roi_poly, (255,))

    image[roi_mask == 0] = 0

    return image, roi_poly, t_matrix


def mask_density(
    mask: np.ndarray,
    roi_mask: np.ndarray = None,
    cell_size: tuple = (5, 5),
    resolution: float = 20
):
    """Compute a density map from a mask image.

    Parameters
    ----------
    mask : array-like of shape = [m, n] and ``dtype=np.uint8``
        Mask image formed by non zero and zero elements.
    roi_mask : array-like of shape = [m, n] and ``dtype=np.uint8``
        Region-of-interest mask image.
    cell_size : tuple, optional, (default=(10, 10))
        Size (width x height) in meters of each rectangular cell of the
        density map grid.
    resolution : float, optional, (default=20)
        Resolution in pixels/meter of the input mask image.

    Returns
    -------
    density_map : array-like of shape = [m, n] and ``dtype=np.float32``
        An array of the same shape as the input `image`, representing the
        density map. To compute the density map, the input image is divided by
        an uniform grid with cell size `cell_size`. Then, inside each cell, the
        density is computed as the ratio between the number non zero pixels
        that belongs to `mask` and the number of non zero pixels from
        `roi_mask`.

    """
    roi_ratio_thr = 0.1
    cell_size = tuple([int(s * resolution) for s in cell_size])

    density_map = np.zeros(mask.shape[0:2], np.float32)

    def set_density(cell: tuple):
        cell_area = roi_area = (cell[1] - cell[0]) * (cell[3] - cell[2])
        if cell_area <= 0:
            return

        if roi_mask is not None:
            roi_area = np.count_nonzero(
                roi_mask[cell[0]:cell[1], cell[2]:cell[3]]
            )

        density = 0
        if (roi_area / cell_area) > roi_ratio_thr:
            density = np.count_nonzero(
                mask[cell[0]:cell[1], cell[2]:cell[3]]
            ) / roi_area
        density_map[cell[0]:cell[1], cell[2]:cell[3]] = density

    h, w = mask.shape[0:2]
    r1 = 0
    while r1 < h:
        c1 = 0
        r2 = min(r1 + cell_size[1], h)
        while c1 < w:
            c2 = min(c1 + cell_size[0], w)
            set_density((r1, r2, c1, c2))
            c1 += cell_size[0]
        r1 += cell_size[1]

    return density_map


def detect_roi(
    veg_mask,
    row_sep: float = 0.7,
    resolution: float = 20,
    min_ratio: float = 0.1,
    max_ratio: float = 0.9
):
    """Detect the region-of-interest in the vegetation mask of a crop field.

    Parameters
    ----------
    veg_mask : array-like of shape = [m, n] and ``dtype=np.uint8``
        Vegetation mask image, where non zero and zero elements represents
        vegetation and background pixels, respectively.
    row_sep : float, optional, (default=0.7)
        Rough estimated of the mean crop row separation.
    resolution : float, optional, (default=20)
        Resolution in pixels/meter of the input mask image.
    min_ratio : float, optional, (default=0.1)
        If the ratio between the area of the detected ROI and the full image
        area if higher than `min_ratio` or lower `max_ratio`, the detected ROI
        is discarded and it is assumed as the whole image.
    max_ratio : float, optional, (default=0.1)
        If the ratio between the area of the detected ROI and the full image
        area if higher than `min_ratio` or lower `max_ratio`, the detected ROI
        is discarded and it is assumed as the whole image.

    Returns
    -------
    roi_poly : array-like of shape = [r, 1, 2] and ``dtype=np.int32``
        Convex polygon that encloses the detected ROI, in the format
        returned by ``help(cv2.contourArea)``.

    """
    dilation = int(2 * row_sep * resolution)
    if dilation > 0:
        se_rect = cv.getStructuringElement(
            cv.MORPH_RECT, (dilation, dilation)
        )
        veg_mask = cv.morphologyEx(veg_mask, cv.MORPH_DILATE, se_rect)

    contours, _ = cv.findContours(
        veg_mask,
        cv.RETR_EXTERNAL,
        cv.CHAIN_APPROX_SIMPLE
    )
    contours_area = list(map(cv.contourArea, contours))
    max_ind = int(np.argmax(contours_area))
    area_ratio = contours_area[max_ind] / veg_mask.size

    if min_ratio < area_ratio < max_ratio:
        roi_poly = cv.convexHull(contours[max_ind])
    else:
        (h, w) = veg_mask.shape
        roi_poly = np.int32(
            [[[0, 0]], [[w, 0]], [[w, h]], [[0, h]]]
        )
    return roi_poly



# -*- coding: utf-8 -*-
from math import pi
from typing import Tuple

import cv2 as cv
import numpy as np
from scipy.signal import convolve

from .utils import transform_matrix


def detect_direction(
    veg_mask: np.ndarray,
    window_shape: Tuple[float, float] = (10, 10),
    n_steps: int = 360,
    resolution: float = 20
):
    """Detect the crop rows direction in a vegetation mask of a crop field.

    The algorithm works as follows. A window is placed in the center of the
    vegetation mask image. This window is rotated in clock wise direction by
    uniformly spaced angles in the interval `[0, pi]`. Inside each window,
    a vegetation profile is computed. A vegetation profile is array where each
    element represents the number of non zero pixels along the corresponding
    column of pixels inside the region of the vegetation mask selected by the
    rotated window. For each vegetation profile, the peak-to-peak variation is
    computed and associated to the rotation angle of the corresponding window.
    The rotated window with the highest peak-to-peak vegetation profile is
    assumed to be perpendicular to the crop rows direction.

    Parameters
    ----------
    veg_mask : array-like of shape = [m, n] and ``dtype=np.uint8``
        Vegetation mask image, where non zero and zero elements represents
        vegetation and background pixels, respectively.
    window_shape : tuple, optional, (default=(200, 200))
        The shape of the window inside which vegetation profiles are computed.
    n_steps : int, optional, (default=360)
        Number of sample angles to rotate the profile window.
    resolution : float, optional, (default=20)
        Resolution in pixels/meter of the input mask image.

    Returns
    -------
    direction : float
        The detected mean crop rows direction.

    """
    max_resolution = 10

    scale = max_resolution / resolution
    if scale < 1:
        veg_mask = cv.resize(
            veg_mask,
            (0, 0),
            fx=scale,
            fy=scale,
            interpolation=cv.INTER_NEAREST
        )
        resolution = 10

    window_shape = (
        int(window_shape[0] * resolution),
        int(window_shape[1] * resolution)
    )

    image_shape = veg_mask.shape[0:2]

    center_point = np.array([
        image_shape[1] / 2,
        image_shape[0] / 2
    ])

    angles = np.linspace(0, pi, n_steps)

    y_grid, x_grid = np.mgrid[
                     0:int(window_shape[0]),
                     0:int(window_shape[1])
                     ]

    window_shape = y_grid.shape

    points = np.zeros((2, y_grid.size), dtype=y_grid.dtype)
    points[0, :] = x_grid.flat
    points[1, :] = y_grid.flat

    scores = np.zeros((angles.size,))
    for angle_ind, angle in enumerate(angles):
        tm = transform_matrix(angle=angle)[:2, :2]
        t_points = np.dot(tm, points)
        t_points = np.add(
            t_points,
            center_point.reshape((-1, 1)),
            dtype=np.int32,
            casting='unsafe'
        )

        t_points[0] = np.clip(t_points[0], a_min=0, a_max=image_shape[1] - 1)
        t_points[1] = np.clip(t_points[1], a_min=0, a_max=image_shape[0] - 1)

        outside = (
                (t_points[0] < 0) |
                (t_points[0] >= image_shape[1]) |
                (t_points[1] < 0) |
                (t_points[1] >= image_shape[0])
        ).reshape(window_shape)

        window = veg_mask[t_points[1], t_points[0]].reshape(window_shape)

        window[outside] = 0
        profile = window.sum(axis=0)
        scores[angle_ind] = profile.ptp()

    best_ind = np.argmax(scores)
    best_angle = angles[best_ind]
    return pi / 2 - best_angle


def detect_rows(
    veg_mask: np.ndarray,
    roi_mask: np.ndarray = None,
    row_sep: float = 0.7,
    extent_max: float = 5,
    extent_thr: float = 0.5,
    fusion_thr: float = 0.4,
    link_thr: int = 3,
    resolution: float = 20
):
    """Detect the crop rows in a vegetation mask of a crop field.

    The algorithm works as follows. A window is placed in the center of the
    vegetation mask image. This window is rotated in clock wise direction by
    uniformly spaced angles in the interval `[0, pi]`. Inside each window,
    a vegetation profile is computed. A vegetation profile is array where each
    element represents the number of non zero pixels along the corresponding
    column of pixels inside the region of the vegetation mask selected by the
    rotated window. For each vegetation profile, the peak-to-peak variation is
    computed and associated to the rotation angle of the corresponding window.
    The rotated window with the highest peak-to-peak vegetation profile is
    assumed to be perpendicular to the crop rows direction.

    Parameters
    ----------
    veg_mask : array-like of shape = [m, n] and ``dtype=np.uint8``
        Vegetation mask image, where non zero and zero elements represents
        vegetation and background pixels, respectively.
    roi_mask : array-like of shape = [m, n] and ``dtype=np.uint8``
        Region-of-interest mask image.
    row_sep : float, optional, (default=0.7)
        Rough estimated of the mean crop row separation.
    extent_max : float, optional, (default=5)
        The maximum length, in meters, of each line of pixels added to eval an
        element of a normalized vegetation profile.
    extent_thr : float, optional, (default=0.5)
        The minimum length, in meters, of each line of pixels added to eval an
        element of a normalized vegetation profile.
    fusion_thr : float, optional, (default=0.4)
        The maximum standard deviation of the row indices of a set of anchor
        points for these points to be fused into a single crop row line.
    link_thr : int, optional, (default=3)
        The minimum number of anchor points required to be fused in a single
        crop row line.
    resolution : float, optional, (default=20)
        Resolution in pixels/meter of the input mask image.

    Returns
    -------
    rows_ridges : array-like of shape = [r, c, 2] and ``dtype=np.int32``
        Detected crop rows lines, ridged aligned, where each element is crop
        row line represented by an array of `c` pairs of indexes in the form
        `(row, column)`. Each crop row line is a continuous polygonal line.
    rows_furrows : array-like of shape = [r, c, 2] and ``dtype=np.int32``
        Detected crop rows lines, furrow aligned, where each element is crop
        row line represented by an array of `c` pairs of indexes in the form
        `(row, column)`. Each crop row line is a continuous polygonal line.

    """
    extent_max = int(extent_max * resolution)
    extent_thr = int(extent_thr * resolution)
    fusion_thr *= resolution
    max_deviation = int(0.5 * row_sep * resolution)
    filter_size = int(row_sep * resolution)
    peaks_mpd = int(0.5 * row_sep * resolution)

    (height, width) = veg_mask.shape
    n_profiles = int(width / extent_max)
    profile_ind = np.linspace(0, width - 1, n_profiles, False, dtype=np.int32)
    profiles_peaks = []

    extent_max *= 255
    extent_thr *= 255

    for ind in profile_ind:
        next_ind = min(width, ind + extent_max)
        distances = extent_max
        if roi_mask is not None:
            distances = roi_mask[:, ind:next_ind].sum(1, np.float32)
            distances[distances < extent_thr] = np.inf
        profile = veg_mask[:, ind:next_ind].sum(1, np.float32) / distances
        profile_peaks = _find_profile_peaks(
            profile,
            filter_size=filter_size,
            mpd=peaks_mpd,
        )
        profiles_peaks.append(profile_peaks)

    linked_peaks = []
    for k in range(n_profiles):
        n_peaks = profiles_peaks[k].size
        for n in range(n_peaks):
            if profiles_peaks[k][n] >= 0:
                rows = _link_peaks(
                    profiles_peaks, k, n, max_deviation, link_thr
                )
                if len(rows) > 0:
                    linked_peaks.append(rows)

    if len(linked_peaks) == 0:
        return np.array([], np.int32), np.array([], np.int32)

    linked_peaks = np.array(linked_peaks)
    linked_peaks = linked_peaks[linked_peaks.mean(1).argsort(), :]
    n_rows = linked_peaks.shape[0]

    linked_peaks_fused = []
    k = 0
    while k < n_rows:
        rows = [linked_peaks[k]]
        n = k + 1
        while (n + 1) < n_rows:
            if linked_peaks[k:(n + 1)].std(0).max() <= fusion_thr:
                rows.append(linked_peaks[n])
            else:
                break
            n += 1
        k = n
        linked_peaks_fused.append(np.array(rows).mean(0))

    linked_peaks = np.array(linked_peaks_fused)
    n_rows = linked_peaks.shape[0]

    ridges = np.empty((n_rows, n_profiles + 1, 2), linked_peaks.dtype)
    ridges[:, :-1, 0] = linked_peaks
    ridges[:, -1, 0] = linked_peaks[:, -1]
    ridges[:, :-1, 1] = profile_ind
    ridges[:, -1, 1] = width - 1

    furrows = np.empty((n_rows + 1, n_profiles + 1, 2), linked_peaks.dtype)
    furrows[1:-1, :, 0] = np.round((ridges[0:-1, :, 0] + ridges[1:, :, 0]) / 2.0)
    furrows[0, :, 0] = np.maximum(0, 2 * furrows[1, :, 0] - furrows[2, :, 0])
    furrows[-1, :, 0] = np.minimum(height - 1, 2 * furrows[-2, :, 0] - furrows[-3, :, 0])
    furrows[:, :, 1] = ridges[0, :, 1]

    return np.int32(ridges), np.int32(furrows)


def _find_profile_peaks(
        profile,
        filter_size: int = 100,
        mpd: int = 50,
        mph: float = 0.2,
        rpl: float = 0.95
):
    mph = mph * profile.max()
    if filter_size > 0:
        profile = convolve(profile / filter_size, np.ones(filter_size), 'same')

    raw_peaks = _detect_peaks(profile, mph=mph, mpd=mpd)

    # Refine peaks locations
    n_peaks = raw_peaks.size
    peaks = np.zeros_like(raw_peaks)

    for k in range(n_peaks):
        this_peak = raw_peaks[k]

        if k != (n_peaks - 1):
            next_peak = raw_peaks[k + 1] + 1
        else:
            next_peak = profile.size

        if k != 0:
            prev_peak = raw_peaks[k - 1]
        else:
            prev_peak = 0

        dx = np.nonzero(
            profile[this_peak:next_peak] < rpl * profile[this_peak]
        )[0]

        if dx.size != 0:
            right_peak = this_peak + dx[0]
        else:
            right_peak = this_peak

        dx = np.nonzero(
            profile[this_peak:prev_peak:-1] < rpl * profile[this_peak]
        )[0]

        if dx.size != 0:
            left_pos = this_peak - dx[0]
        else:
            left_pos = this_peak

        peaks[k] = round((left_pos + right_peak) / 2)

    return peaks


def _detect_peaks(
        x: np.ndarray,
        mph: float = None,
        mpd: float = 1,
        threshold: float = 0,
        edge: str = 'rising',
        kpsh: bool = False
):
    """Detect peaks in data based on their amplitude and other features.

    Parameters
    ----------
    x : 1D array_like
        data.
    mph : {None, number}, optional (default = None)
        detect peaks that are greater than minimum peak height.
    mpd : positive integer, optional (default = 1)
        detect peaks that are at least separated by minimum peak distance (in
        number of data).
    threshold : positive number, optional (default = 0)
        detect peaks that are greater than `threshold`
        in relation to their immediate neighbors.
    edge : {None, 'rising', 'falling', 'both'}, optional (default = 'rising')
        for a flat peak, keep only the rising edge ('rising'), only the
        falling edge ('falling'), both edges ('both'), or don't detect a
        flat peak (None).
    kpsh : bool, optional (default = False)
        keep peaks with same height even if they are closer than `mpd`.

    Returns
    -------
    ind : 1D array_like
        indexes of the peaks in `x`.

    The function can handle NaN's
    """

    x = np.atleast_1d(x).astype('float64')
    if x.size < 3:
        return np.array([], dtype=int)

    # find indices of all peaks
    dx = x[1:] - x[:-1]
    # handle NaN's
    indnan = np.where(np.isnan(x))[0]
    if indnan.size:
        x[indnan] = np.inf
        dx[np.where(np.isnan(dx))[0]] = np.inf
    ine, ire, ife = np.array([[], [], []], dtype=int)
    if not edge:
        ine = np.where((np.hstack((dx, 0)) < 0) & (np.hstack((0, dx)) > 0))[0]
    else:
        if edge.lower() in ['rising', 'both']:
            ire = np.where(
                (np.hstack((dx, 0)) <= 0) & (np.hstack((0, dx)) > 0)
            )[0]
        if edge.lower() in ['falling', 'both']:
            ife = np.where(
                (np.hstack((dx, 0)) < 0) & (np.hstack((0, dx)) >= 0)
            )[0]
    ind = np.unique(np.hstack((ine, ire, ife)))
    # handle NaN's
    if ind.size and indnan.size:
        # NaN's and values close to NaN's cannot be peaks
        ind = ind[np.in1d(ind, np.unique(
            np.hstack((indnan, indnan - 1, indnan + 1))
        ), invert=True)]
    # first and last values of x cannot be peaks
    if ind.size and ind[0] == 0:
        ind = ind[1:]
    if ind.size and ind[-1] == x.size - 1:
        ind = ind[:-1]
    # remove peaks < minimum peak height
    if ind.size and mph is not None:
        ind = ind[x[ind] >= mph]
    # remove peaks - neighbors < threshold
    if ind.size and threshold > 0:
        dx = np.min(
            np.vstack([x[ind] - x[ind - 1], x[ind] - x[ind + 1]]),
            axis=0
        )
        ind = np.delete(ind, np.where(dx < threshold)[0])
    # detect small peaks closer than minimum peak distance
    if ind.size and mpd > 1:
        ind = ind[np.argsort(x[ind])][::-1]  # sort ind by peak height
        idel = np.zeros(ind.size, dtype=bool)
        for i in range(ind.size):
            if not idel[i]:
                # keep peaks with the same height if kpsh is True
                idel = idel | (ind >= ind[i] - mpd) & (ind <= ind[i] + mpd) \
                       & (x[ind[i]] > x[ind] if kpsh else True)
                idel[i] = 0  # Keep current peak
        # remove the small peaks and sort back the indices by their occurrence
        ind = np.sort(ind[~idel])

    return ind


def _link_peaks(profile_peaks, profile_ind, peak_ind, max_dev, link_thr=1):
    n_profiles = len(profile_peaks)
    linked_peaks = np.empty(n_profiles, np.int32)
    linked_peaks[0:(profile_ind + 1)] = profile_peaks[profile_ind][peak_ind]
    profile_peaks[profile_ind][peak_ind] = -(max_dev + 1)
    n_linked = 1
    for k in range(profile_ind + 1, n_profiles):
        if profile_peaks[k].size == 0:
            linked_peaks[k] = linked_peaks[k - 1]
            continue
        dev = np.abs(profile_peaks[k] - linked_peaks[k - 1])
        min_ind = dev.argmin()
        if dev[min_ind] < max_dev:
            linked_peaks[k] = profile_peaks[k][min_ind]
            profile_peaks[k][min_ind] = -(max_dev + 1)
            n_linked += 1
        else:
            linked_peaks[k] = linked_peaks[k - 1]

    if n_linked >= link_thr:
        return linked_peaks
    return []